#!/usr/local/bin/python3.6
#####################################################################################
###
### Acme Software Works, Inc.
###
### Created June 16, 2019 by Don Dugger
###
### <PRE>
### Copyright 2019 Acme Software Works, Inc.
### 
### Redistribution and use in source and binary forms, with or without
### modification, are permitted provided that the following conditions are met:
### 
### 1. Redistributions of source code must retain the above copyright notice,
###    this list of conditions and the following disclaimer.
### 
### 2. Redistributions in binary form must reproduce the above copyright notice,
###    this list of conditions and the following disclaimer in the documentation
###    and/or other materials provided with the distribution.
### 
### THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
### AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
### IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
### ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
### LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
### CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
### GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
### HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
### LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
### OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
### <PRE>
#####################################################################################
from PyQt5 import QtWidgets, uic, QtCore, QtGui
import time
from time import gmtime, strftime
from threading import Timer
from datetime import datetime
from dateutil.relativedelta import relativedelta
import getopt, sys, os
from promptdb import PromptDB
from promptdialog import PromptDialog
from libprompt import monthDay
from libprompt import UI_DIR
from libprompt import PROMPT_DATE_FORMAT
import re


#####################################################################################

def usage():
    print("Usage: {} <-i id (prompt db id)>".format(os.path.basename(sys.argv[0])))

#####################################################################################
#
# The class Main is the program. The functional part that is outside Main is getting
# the command line option. 
#
class Main(QtWidgets.QMainWindow):
    def __init__(self):
        super(Main, self).__init__()
        uic.loadUi(UI_DIR + 'prompt.ui', self)

        # Get the prompt dialog box for editing.
        self.dlg = PromptDialog()

        # Retrieve and display the date, time and task text.
        self.db = PromptDB()
        self.prompt = self.db.fetchOnePrompt(prompt_id)
        self.prompt_datetime = datetime.strptime(str(self.prompt[1]) + " " + str(self.prompt[2]),"%Y-%m-%d %H:%M:%S")
        self.dateText.setText(self.prompt_datetime.strftime(PROMPT_DATE_FORMAT))
        self.taskText.setText(self.prompt[4])

        # Connect the 'Later' button to exit() and make it
        # the default. Making it the fault is so the user
        # doesn't accidentally mark the task complete.
        self.laterButton.clicked.connect(exit)
        self.laterButton.setAutoDefault(True)
        # Connect the rest of the buttons to there respective methods. 
        self.editButton.clicked.connect(self._edit)
        self.completeButton.clicked.connect(self._complete)

        # The prompt has a 3 minute time out so that
        # prompt don't accumulate on the user's screen.
        # The time has a progress bar to show the user how much
        # time is left. 
        self.show()
        self.clock = 180 # 3 Minutes
        self.timer = QtCore.QTimer(self)
        self.timer.timeout.connect(self._tic)
        self.timer.start(3000)

    # The progress bar is updated every 3 seconds.
    def _tic(self):
        if self.clock > 0:
            self.clock -= 3
            self.progressBar.setValue(self.clock)
        else:
            exit(0)

    # To edit just load the prompt dialog box with the info that was fetched by the constructor.
    def _edit(self):
        # First change the time from 24 hour to 12 hour plus the period (AM/PM).
        prompt_time = datetime.strptime(str(self.prompt[2]),"%H:%M:%S")
        # Then display the prompt dialog box.
        self.dlg.edit(prompt_id,self.prompt[1].strftime("%Y-%m-%d"),prompt_time.strftime("%I:%M %p"),self.prompt[3],self.prompt[4])

    # When a task is complete it's entree in the database needs to be updated. If the frequency is 'Once' the record will be
    # deleted otherwise the due date will be updated in accordance with the frequency spec.
    # Note that the frequency spec was validated upon the tasks entree and will not be re-validated here.
    def _complete(self):
        # First break a part the frequency spec.
        freq_list = re.split(',',self.prompt[3])
        x = 1 # If there is no multiplier the frequency is once.
        mult_list = re.split('\*',freq_list[0])
        if len(mult_list) > 1: # Is there a multiplier?
            # The second item in the list is the multiplier.
            x = int(mult_list[1])
        # Deal with limits
        count = int(self.prompt[5])
        # Note if the count starts at 0 the count is not active.
        if count: # Is this task being counted down?
            count -= 1
            if count == 0: # If it's finished delete it.
                self.db.delPrompt(prompt_id)
            else: # Update the database record.
                self.db.setCount(prompt_id,count)
        # Now process main frequency spec
        if re.match("once$",freq_list[0],re.IGNORECASE): # Delete & exit.
            self.db.delPrompt(prompt_id)
            exit(0)
        elif re.match("daily($|\*[0-9]+$)",freq_list[0],re.IGNORECASE): # Update to days times x.
            self.prompt_datetime += relativedelta(days=+x)
        elif re.match("weekly($|\*[0-9]+$)",freq_list[0],re.IGNORECASE): # Update to weeks times x.
            self.prompt_datetime += relativedelta(weeks=+x)
        elif re.match("monthly($|\*[0-9]+$)",freq_list[0],re.IGNORECASE): # Update to months times x.
            # If there is a modifier and it not the limit modifier update accordingly.
            if len(freq_list) > 1 and not re.match("limit=[0-9]+$",freq_list[1],re.IGNORECASE):
                # Get the day from the prompt.
                day = int(self.prompt_datetime.strftime("%e"))
                # Move forward x months.
                self.prompt_datetime += relativedelta(months=+x)
                # Get the day of the month-day spec.
                the_day = monthDay(self.prompt_datetime,freq_list[1])
                # Calculate the proper new day.
                self.prompt_datetime += relativedelta(days=+(the_day-day))
            else: # If there is no modifier simply move ahead x months.
                self.prompt_datetime += relativedelta(months=+x)
        elif re.match("yearly($|\*[0-9]+$)",freq_list[0],re.IGNORECASE): # Update to years times x.
            self.prompt_datetime += relativedelta(years=+x)
        self.db.setDate(prompt_id,self.prompt_datetime.strftime("%Y-%m-%d"))
        exit(0)

#####################################################################################
################ Main ###############################################################
#####################################################################################
# The propmpt.py program displays the 'task' that is due. 
# It requires the database id in order to retrieve the 'task' information.
# It does not determine if the 'task' is due it simply displays it. It will
# update completed 'tasks'.
#
# Start by checking that the id has been entered correctly.
try:
    opts, args = getopt.getopt(sys.argv[1:], "i:")
except getopt.GetoptError as err:
    print(str(err))
    usage()
    sys.exit(2)

# The id is needed globally.
global prompt_id 
prompt_id = -1 # Set it to a invalid value to test later.
for opt, arg in opts:
    if opt == "-i":
        prompt_id = int(arg)
    else:
        usage()
        sys.exit(2)

if prompt_id < 0: # Was a valid value entered?
    usage()
    sys.exit(0)

# Start everything up.
app = QtWidgets.QApplication(sys.argv)
window = Main()
sys.exit(app.exec())

            
