#####################################################################################
###
### Acme Software Works, Inc.
###
### Created June 16, 2019 by Don Dugger
###
### <PRE>
### Copyright 2019 Acme Software Works, Inc.
### 
### Redistribution and use in source and binary forms, with or without
### modification, are permitted provided that the following conditions are met:
### 
### 1. Redistributions of source code must retain the above copyright notice,
###    this list of conditions and the following disclaimer.
### 
### 2. Redistributions in binary form must reproduce the above copyright notice,
###    this list of conditions and the following disclaimer in the documentation
###    and/or other materials provided with the distribution.
### 
### THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
### AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
### IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
### ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
### LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
### CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
### GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
### HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
### LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
### OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
### <PRE>
#####################################################################################
from PyQt5 import QtWidgets, uic, QtCore, QtGui
from PyQt5.QtCore import QDate, QTime, QDateTime, Qt, QRect, QPoint
from PyQt5.QtWidgets import QMessageBox
from datetime import datetime
from dateutil.relativedelta import relativedelta
import getopt, sys
from promptdb import PromptDB
from libprompt import monthDay
from libprompt import UI_DIR
import re

#####################################################################################
# The Frequency Combo Box.
# This is one of the primary features of 'TaskPrompt'.
# The FreqBox is a separate class (object) in order to
# validate the frequency specification during entree.
# By deriving it from the base class QComboBox. The 
# method 'focusOutEvent()' can be replaced so the frequency
# specification can immediately validated. 
class FreqBox(QtWidgets.QComboBox):
    def __init__(self,parent,x,y): # Let the caller set the position. 
        super(FreqBox, self).__init__(parent)
        self.setGeometry(QRect(x,y,250,30))
        # Generally set up the appearance. 
        font = QtGui.QFont()
        font.setFamily("Sans Serif")
        font.setPointSize(11)
        font.setItalic(False)
        self.setFont(font)
        self.setFocusPolicy(Qt.StrongFocus)
        # Load the basic frequencies.
        freq_list = [
            'Once',
            'Daily',
            'Weekly',
            'Monthly',
            'Yearly',
            ]
        self.clear()
        self.addItems(freq_list)
        # It is editable so modifier can be added.
        self.setEditable(True)

        self.show()

    # Validate the specification.
    def isFreqValid(self):
        freq_list = re.split(',',self.currentText())
        # The first argument is the frequency.
        if not re.match("once$|daily($|\*[0-9]+$)|weekly($|\*[0-9]+$)|monthly($|\*[0-9]+$)|yearly($|\*[0-9]+$)",freq_list[0],re.IGNORECASE):
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Critical)
            msg.setText("Incorrect frequency specification!")
            msg.exec()
            return False
        # The second modifiers can be a day modifier or a limit modifier.
        if len(freq_list) > 1:
            if not (re.match("monthly",freq_list[0],re.IGNORECASE)\
               and re.match(".+sun$|.+mon$|.+tue$|.+wed$|.+thu$|.+fri$|.+sat$|.+day$",freq_list[1],re.IGNORECASE)\
               and re.match("^first...$|^second...$|^third...$|^fourth...$|last...$",freq_list[1],re.IGNORECASE)\
               or re.match("limit=[0-9]+$",freq_list[1],re.IGNORECASE)):
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Critical)
                msg.setText("Incorrect frequency modifier specification!")
                msg.exec()
                return False
        # The third modifier can only be a limit modifier.
        if len(freq_list) > 2:
            if not re.match("limit=[0-9]+$",freq_list[2],re.IGNORECASE):
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Critical)
                msg.setText("Incorrect number modifiers!")
                msg.exec()
                return False
        # Number is 1 frequency 1 day modifiers plus a limit modifier for a total of 3.
        if len(freq_list) > 3:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Critical)
            msg.setText("Incorrect number modifiers!")
            msg.exec()
            return False
        return True

    # This replaces the base class method so the focus
    # event will come here. And when the focus is lost
    # it is validated.
    def focusOutEvent(self, event):
        if event.lostFocus():
            self.isFreqValid()
        # Now let the base class have the event.
        super(FreqBox, self).focusOutEvent(event)
        

#####################################################################################
# The Prompt Dialog Box
#
# This dialog box is used for adding and editing prompts.
#
class PromptDialog(QtWidgets.QDialog):
    def __init__(self):
        super(PromptDialog, self).__init__()
        # Setup the Prompt entry/edit dialog box
        uic.loadUi(UI_DIR + 'taskprompt-input.ui',self)
        self.datepicker = uic.loadUi(UI_DIR + 'datepicker.ui')
        self.help = uic.loadUi(UI_DIR + 'prompthelp.ui')

        # Setup the 'datepicker'.
        self.datepicker.calendar.activated.connect(self._updateDate)
        self.datepicker.buttonBox.accepted.connect(self._updateDateOK)
        self.datepicker.buttonBox.rejected.connect(self.datepicker.close)
        self.datepicker.todayButton.clicked.connect(self.setToday)
        self.dateButton.clicked.connect(self.datepicker.exec)
        # Connect the help screen.
        self.buttonBox.helpRequested.connect(self.help.exec)
        self.help.buttonBox.accepted.connect(self.help.close)
        # Init strings for edit change test.
        self.date_str = ""
        self.time_str = ""
        self.freq_str = ""

        # The time combo box.
        # The list of times.
        time_list = [
            '12:00 AM', '12:15 AM', '12:30 AM', '12:45 AM', '01:00 AM', '01:15 AM', '01:30 AM', '01:45 AM',
            '02:00 AM', '02:15 AM', '02:30 AM', '02:45 AM', '03:00 AM', '03:15 AM', '03:30 AM', '03:45 AM',
            '04:00 AM', '04:15 AM', '04:30 AM', '04:45 AM', '05:00 AM', '05:15 AM', '05:30 AM', '05:45 AM',
            '06:00 AM', '06:15 AM', '06:30 AM', '06:45 AM', '07:00 AM', '07:15 AM', '07:30 AM', '07:45 AM',
            '08:00 AM', '08:15 AM', '08:30 AM', '08:45 AM', '09:00 AM', '09:15 AM', '09:30 AM', '09:45 AM',
            '10:00 AM', '10:15 AM', '10:30 AM', '10:45 AM', '11:00 AM', '11:15 AM', '11:30 AM', '11:45 AM',
            '12:00 PM', '12:15 PM', '12:30 PM', '12:45 PM', '01:00 PM', '01:15 PM', '01:30 PM', '01:45 PM',
            '02:00 PM', '02:15 PM', '02:30 PM', '02:45 PM', '03:00 PM', '03:15 PM', '03:30 PM', '03:45 PM',
            '04:00 PM', '04:15 PM', '04:30 PM', '04:45 PM', '05:00 PM', '05:15 PM', '05:30 PM', '05:45 PM',
            '06:00 PM', '06:15 PM', '06:30 PM', '06:45 PM', '07:00 PM', '07:15 PM', '07:30 PM', '07:45 PM',
            '08:00 PM', '08:15 PM', '08:30 PM', '08:45 PM', '09:00 PM', '09:15 PM', '09:30 PM', '09:45 PM',
            '10:00 PM', '10:15 PM', '10:30 PM', '10:45 PM', '11:00 PM', '11:15 PM', '11:30 PM', '11:45 PM',
            ]
        # Clear and initialize the time combo box.
        self.timeBox.clear()
        self.timeBox.addItems(time_list)
        # Force the 'timeBox' to obey the drop down limit (maxVisibleItems).
        # When 'QComboBox'es are set to not editable it is ignored. 
        self.timeBox.setStyleSheet("QComboBox { combobox-popup: 0; }");

        # In order to put the 'FreqBox' in a place
        # based on the layout produced by Qt Designer
        # it is placed in reference to the frequency 
        # label 'label_freq'.
        x = self.label_freq.pos().x()
        y = self.label_freq.pos().y()
        w = self.label_freq.size().width()
        # Put 'freqBox' behind the label plus 10.
        self.freqBox = FreqBox(self,x+w+10,y)
        self.buttonBox.accepted.connect(self._doAccept)
        # With the 'freqBox' being setup programmatically
        # the tab order will need to set.
        self.setTabOrder(self.dateBox,self.dateButton)
        self.setTabOrder(self.dateButton,self.timeBox)
        self.setTabOrder(self.timeBox,self.freqBox)
        self.setTabOrder(self.freqBox,self.taskText)
        # Initialize the edit mode state machine.
        self.state = None
        self.EDIT = 1
        self.ADD = 2
        # Set the add id to unknown.
        self.add_id = -1

    # Setting the date picker to today's date. This implements the
    # 'datepicker's 'Today' button. 
    def setToday(self):
        self.datepicker.calendar.setSelectedDate(QDate.currentDate())

    # When the accept ( The "OK" button ) is received determine
    # which mode the dialog box was started in.
    def _doAccept(self):
        if self.state == self.EDIT:
            self._acceptEdit()
        elif self.state == self.ADD:
            self._acceptAdd()
        self.close()

    # When a date is chosen in the 'calendar' object
    # of the 'datepicker' update the 'dateBox' text.
    def _updateDate(self, date):
        self.dateBox.setText(date.toString("yyyy-MM-dd"))
        self.datepicker.close()

    # When a date is chosen in the 'datapicker' it is 
    # validated and the 'dateBox' text is updated.
    def _updateDateOK(self):
        date = self.datepicker.calendar.selectedDate()
        self.dateBox.setText(date.toString("yyyy-MM-dd"))
        now = datetime.now()
        if now >= self._getDateTime():
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Critical)
            msg.setText("The date must be in the future")
            msg.exec()
            return
        self._updateDate(date)

    # With the dialog box's 'dateBox' and 'timeBox' retrive them
    # convert to a 'datetime' object and return it.
    def _getDateTime(self):
        # Retrieve the date which is a string. 
        date_str = self.dateBox.text()
        # Split into year, month and day strings which
        # can the be converted to integers.
        date_list = date_str.split('-')
        # The time is in a 12 hour format it needs
        # to be in a 24 hour format.
        time_str = self.timeBox.currentText()
        # Split it up into hours, minutes and period (AM/PM).
        time_list1 = time_str.split(' ')
        time_list2 = time_list1[0].split(':')
        # Convert hours and minutes string to integers.
        hr = int(time_list2[0])
        mn = int(time_list2[1])
        # Now deal with period.
        if hr == 12 and time_list1[1] == "AM":
            hr = 0
        elif time_list1[1] == "PM" and hr != 12:
            hr += 12
        # Return the 'datetime' object.
        return datetime(int(date_list[0]),int(date_list[1]),int(date_list[2]),hr,mn,0)

    # When in ADD mode and the 'OK' button is pressed
    # this is the method that is invoked.
    def _acceptAdd(self):
        # Retrieve the date and time.
        the_date = self._getDateTime()
        now = datetime.now()
        # Make sure the date be entered is in the future.
        if now > the_date:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Critical)
            msg.setText("The date must be in the future")
            msg.exec()
            return
        # Get a database connection.
        db = PromptDB()
        # If no limit is set the count is initialized to zero.
        count = 0
        # If 'Monthly' is set check for modifiers.
        if re.match("monthly",self.freqBox.currentText(),re.IGNORECASE):
            freq_str = self.freqBox.currentText()
            freq_list = freq_str.split(',')
            # If the first modifier is not a 'limit' modifier it's 
            # a day of the month modifier. 
            # Note that frequency specification has already been validated.
            if not re.match("limit=[0-9]+",freq_list[1],re.IGNORECASE):
                # The start date will now be set to the first due date.
                # First get the modified day of the month.
                the_day = monthDay(the_date,freq_list[1])
                # The day entered.
                day = int(the_date.strftime("%e"))
                # Adjust the 'datetime' object to the due date.
                if the_day <= day:
                    the_date += relativedelta(months=+1)
                    the_day = monthDay(the_date,freq_list[1])
                the_date += relativedelta(days=+(the_day-day))
        # Has a limit been set.
        if re.match(".+limit=[0-9]+$",self.freqBox.currentText(),re.IGNORECASE):
            freq_list = self.freqBox.currentText().split('=')
            count = int(freq_list[1]) # Set the count.
        # Add the new prompt to the database and set the add id so the caller can retrieve it.
        self.add_id = db.addPrompt(the_date.strftime("%Y-%m-%d"),
                                   the_date.strftime("%H:%M:%S"),
                                   self.freqBox.currentText(),
                                   self.taskText.text(),
                                   count)

    # When in EDIT mode and the 'OK' button is pressed
    # this is the method that is invoked.
    def _acceptEdit(self):
        # Retrieve the dialog box's entrees.
        the_date = self._getDateTime()
        date_str = the_date.strftime("%Y-%m-%d")
        time_str = the_date.strftime("%H:%M:%S")
        freq_str = self.freqBox.currentText()
        # Get a database connection.
        db = PromptDB()
        # Initialize the count variable. 
        count = 0
        # If the date or time have been changed check that it is in the future.
        if self.date_str != date_str or self.time_str != time_str:
            now = datetime.now()
            if now > the_date:
                msg = QMessageBox()
                msg.setIcon(QMessageBox.Critical)
                msg.setText("The date must be in the future")
                msg.exec()
                return
        # If the frequency spec has changed deal with it.
        if self.freq_str != freq_str:
            # If the frequency is monthly check for modifiers.
            if re.match("monthly",self.freqBox.currentText(),re.IGNORECASE):
                freq_str = self.freqBox.currentText()
                freq_list = freq_str.split(',')
                # If the modifier is not a limit update the due date
                if not re.match("limit=[0-9]+",freq_list[1],re.IGNORECASE):
                    the_day = monthDay(the_date,freq_list[1])
                    day = int(the_date.strftime("%e"))
                    if the_day <= day:
                        the_date += relativedelta(months=+1)
                        the_day = monthDay(the_date,freq_list[1])
                    the_date += relativedelta(days=+(the_day-day))
            # There is limit modifiers update the count.
            if re.match(".+limit=[0-9]+$",self.freqBox.currentText(),re.IGNORECASE):
                freq_list = self.freqBox.currentText().split('=')
                count = int(freq_list[1])
            db.setCount(self.current_id,count)
        # Update the database.
        db.editPrompt(self.current_id,
                      date_str,
                      time_str,
                      freq_str,
                      self.taskText.text())


    # Caller's access to id of the added prompts.
    def getAddId(self):
        return self.add_id

    # This is the caller's access to the edit functionality.
    def edit(self,this_id,date_str,time_str,freq_str,task_str):
        # Save the starting values so changes can be detacted.
        self.date_str = date_str
        self.time_str = time_str
        self.freq_str = freq_str
        # Set the date picker to the date of the prompt.
        # Note Qt has a different date class used in the 'calendar' object.
        date_list = date_str.split('-')
        self.datepicker.calendar.setSelectedDate(QDate(int(date_list[0]),int(date_list[1]),int(date_list[2])))
        self.dateBox.setText(date_str) # The text is just a string
        # Set the time to the time in the prompt
        index = self.timeBox.findText(time_str)
        self.timeBox.setCurrentIndex(index)
        # Set the frequency to the frequency in the prompt
        index = self.freqBox.findText(freq_str)
        if index >= 0:
            self.freqBox.setCurrentIndex(index)
        else:
            self.freqBox.addItem(freq_str)
            index = self.freqBox.count()-1
            self.freqBox.setCurrentIndex(index)
        self.taskText.setText(task_str)
        # Change the return to edit
        self.current_id = this_id
        self.state = self.EDIT
        self.exec()

    # This is the caller's access to the add functionality.
    def add(self):
        today = datetime.today()
        # Calculate the next 15 minute segment
        index = int(today.strftime("%H"))*60
        index += int(today.strftime("%M"))
        index = (index/15)+1
        self.timeBox.setCurrentIndex(index)
        # Set the date picker to today
        self.datepicker.calendar.setSelectedDate(QDate.currentDate())
        self.dateBox.setText(today.strftime("%Y-%m-%d"))
        # Set the frequency to "Once"
        self.freqBox.setCurrentIndex(0)
        self.taskText.clear()
        # Change the return to add
        self.state = self.ADD
        self.exec()

