#!/usr/local/bin/python3.6
#####################################################################################
###
### Acme Software Works, Inc.
###
### Created June 16, 2019 by Don Dugger
###
### <PRE>
### Copyright 2019 Acme Software Works, Inc.
### 
### Redistribution and use in source and binary forms, with or without
### modification, are permitted provided that the following conditions are met:
### 
### 1. Redistributions of source code must retain the above copyright notice,
###    this list of conditions and the following disclaimer.
### 
### 2. Redistributions in binary form must reproduce the above copyright notice,
###    this list of conditions and the following disclaimer in the documentation
###    and/or other materials provided with the distribution.
### 
### THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
### AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
### IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
### ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
### LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
### CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
### GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
### HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
### LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
### OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
### <PRE>
#####################################################################################
from PyQt5 import QtWidgets, uic, QtCore, QtGui
from PyQt5.QtCore import QDate, QTime, QDateTime, Qt
from PyQt5.QtWidgets import QInputDialog,\
                            QLineEdit,\
                            QLabel,\
                            QProgressDialog,\
                            QWidget,\
                            QTableWidgetItem,\
                            QTableWidgetSelectionRange,\
                            QMessageBox
from datetime import datetime
from threading import Timer
from promptdb import PromptDB
from promptdialog import PromptDialog
from libprompt import BIN_DIR
import sys
import subprocess

#####################################################################################
#
# This is the prime user interface to the 'TasksPrompt' system.
#
# It list all pending tasks and allows the user to add, delete and edit them as well
# as checking for currently due ones.
#
class Main(QtWidgets.QMainWindow):
    def __init__(self):
        super(Main, self).__init__()
        uic.loadUi('taskprompt-main.ui', self)
        self.dlg = PromptDialog()

        # Connect the Buttons
        self.quitButton.clicked.connect(self._quit)
        self.quitButton.setAutoDefault(True)
        self.editButton.clicked.connect(self._edit)
        self.deleteButton.clicked.connect(self._delete)
        self.addButton.clicked.connect(self._add)
        self.updateButton.clicked.connect(self._update)

        # Set the column Headers.
        self.promptTable.setHorizontalHeaderLabels(['','Date','Time','Frequency','CT','Task'])
        # Set column widths.
        self.promptTable.setRowHidden(0,True)
        self.promptTable.setColumnWidth(0,0)
        self.promptTable.setColumnWidth(1,90)
        self.promptTable.setColumnWidth(2,80)
        self.promptTable.setColumnWidth(3,185)
        self.promptTable.setColumnWidth(4,25)
        # Set full row selection.
        col_ct = self.promptTable.columnCount()
        self.promptTable.setRangeSelected(QTableWidgetSelectionRange(0, col_ct-1, 0, 1), True)
        # Get the tasks from the datadase.
        self.db = PromptDB()
        self.loadTable()
        # Show the window.
        self.show()
        # Now setup a one minute timer to
        # update prompts that may have changed 
        self.timer = QtCore.QTimer(self)
        self.timer.timeout.connect(self._tic)
        self.timer.start(60000)

    # The timer
    def _tic(self):
        # Update the list
        row = self.promptTable.currentRow()
        self.loadTable()
        self.promptTable.selectRow(row)

    # Connect to the 'Quit' button.
    def _quit(self):
        exit(0)

    # Update the prompt list.
    def loadTable(self):
        # get a fresh connection
        self.db.close()
        self.db.setup()
        # Get the list from the database.
        prompt_list = self.db.fetchAllPrompts()
        self.db.close()
        # Clear the list before re-populating it.
        while self.promptTable.rowCount() > 0:
            self.promptTable.removeRow(0)
        # Go through the list a row at a time.
        for row in prompt_list:
            # If the count (6th column) is zero do not display it.
            count = int(row[5])
            if int(row[5]) > 0:
                count_str = str(count)
            else:
                count_str = ""
            # The time needs to be converted from 24 clock to 12 hr plus the period (AM/PM).
            x = datetime.strptime(str(row[2]), '%H:%M:%S')
            # Get the position into which to place the next row, at the end of the list.
            rowPosition = self.promptTable.rowCount()
            self.promptTable.insertRow(rowPosition)
            # Populate it.
            self.promptTable.setItem(rowPosition , 0, QTableWidgetItem(str(row[0])))
            self.promptTable.setItem(rowPosition , 1, QTableWidgetItem(row[1].strftime("%Y-%m-%d")))
            self.promptTable.setItem(rowPosition , 2, QTableWidgetItem(x.strftime("%I:%M %p")))
            self.promptTable.setItem(rowPosition , 3, QTableWidgetItem(row[3]))
            self.promptTable.setItem(rowPosition , 4, QTableWidgetItem(count_str))
            self.promptTable.setItem(rowPosition , 5, QTableWidgetItem(row[4]))
        # Place the selection at the top.
        self.promptTable.clearSelection()
        self.promptTable.selectRow(0)

    # Connected to the 'Edit' button.
    # Load the dialog box with the selected prompt and display it.
    def _edit(self):
        # Get the selected row.
        row = self.promptTable.currentRow()
        # Get the database id.
        id = self.promptTable.item(row, 0).text()
        # Get the prompt data as strings.
        date_str = self.promptTable.item(row, 1).text()
        time_str = self.promptTable.item(row, 2).text()
        freq_str = self.promptTable.item(row, 3).text()
        task_str = self.promptTable.item(row, 5).text()
        # Hand the dialog box the data. The dialog class will
        # handle verifying and entering the changes in the database.
        self.dlg.edit(id,date_str,time_str,freq_str,task_str)
        # Now update the list.
        self.loadTable()
        # Put the selection back to where it was.
        self.promptTable.selectRow(row)

    # Connected to the 'Delete' button.
    # Delete the prompt from the database and the displayed list.
    def _delete(self):
        # Get the selected prompt.
        row = self.promptTable.currentRow()
        # get a fresh connection
        self.db.close()
        self.db.setup()
        # Delete it from the database.
        self.db.delPrompt(self.promptTable.item(row, 0).text())
        self.db.close()
        # Set the selection near where it was.
        self.promptTable.selectRow(row)
        # And remove it from the list.
        self.promptTable.removeRow(row)

    # Connected to the 'Add' button.
    def _add(self):
        # Display a blank dialog box.
        self.dlg.add()
        # Get the database id of the last entree.
        prompt_id = self.dlg.getAddId()
        # Re-load the list.
        self.loadTable()
        # Find the new prompt if there is one and select it.
        for row in range(self.promptTable.rowCount()):
            if int(self.promptTable.item(row,0).text()) == prompt_id:
                self.promptTable.selectRow(row)

    # Connected to the 'Update' button.
    # This call the same program the cron does. It allows
    # the user to check for due tasks between cron checks.
    def _update(self):
        # Get the current selection so it can be re-display afterwards.
        row = self.promptTable.currentRow()
        # Set the command.
        args = [BIN_DIR+'cronprompt.py']
        # Execute it.
        res = subprocess.run(args, stdout=subprocess.PIPE)
        # Update the list.
        self.loadTable()
        self.promptTable.selectRow(row)

################ Main ######################################################################

app = QtWidgets.QApplication(sys.argv)
window = Main()
sys.exit(app.exec())

